class BaseLexicon(object):

    @staticmethod
    def transaction_begin_query():
        return 'BEGIN;'

    @staticmethod
    def table_schema_query(table_name): pass

    @staticmethod
    def table_name_default(row): pass

    @staticmethod
    def extrapolator(field):
        return '%(' + '%s' % field + ')s'

    @staticmethod
    def update_values(fields, value_extrapolators):
        values = []
        for field_ext in zip(fields, value_extrapolators[0]):
            values += ['%s = %s' % field_ext]
        return ', '.join(values)

    @staticmethod
    def row_ids(db): pass

    @staticmethod
    def apply_watermark(query, watermark):
        return ' '.join([query, watermark])

    @staticmethod
    def format_args(args):
        return args


import sys
if sys.version_info[0] < 3:
    import MySQLdb
else:
    import pymysql as MySQLdb

from jardin.tools import retry
from jardin.database.base import BaseConnection
from jardin.database.lexicon import BaseLexicon


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "SHOW COLUMNS FROM %s;" % table_name

    @staticmethod
    def column_name_default(row):
        return row['Field'], row['Default']

    @staticmethod
    def row_ids(db, primary_key):
        db.execute('SELECT LAST_INSERT_ID();')
        return [db.cursor().fetchall()[0][0]]

    @staticmethod
    def apply_watermark(query, watermark):
        return ' '.join([watermark, query])


class DatabaseConnection(BaseConnection):

    DRIVER = MySQLdb
    LEXICON = Lexicon

    @retry(DRIVER.OperationalError, tries=3)
    def connect(self):
        return super(DatabaseConnection, self).connect()

    @retry(DRIVER.InterfaceError, tries=3)
    def execute(self, *query):
        return super(DatabaseConnection, self).execute(*query)



__all__ = (
    'is_not_null', 'greater_than', 'gt', 'greater_or_equal', 'get',
    'less_than', 'lt', 'less_or_equal', 'let'
    )

def is_not_null():
    return 'IS NOT NULL'

def operator(op, *args, **kwargs):
    def func():
        if len(args) > 0:
            return op, args[0]
        if 'column' in kwargs:
            return '%s %s' % (op, kwargs['field']), None
    return func

def greater_than(*args, **kwargs):
    return operator('>', *args, **kwargs)
gt = greater_than

def greater_or_equal(field):
    return operator('>=', field)
get = greater_or_equal

def less_than(field):
    return operator('<', field)
lt = less_than

def less_or_equal(field):
    return operator('<=', field)
let = less_or_equal


import sqlite3
from memoized_property import memoized_property

from jardin.database.base import BaseConnection
from jardin.database.lexicon import BaseLexicon


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "pragma table_info(%s);" % table_name

    @staticmethod
    def transaction_begin_query():
        return 'BEGIN;'

    @staticmethod
    def column_name_default(row):
        return row['name'], row['dflt_value']

    @staticmethod
    def extrapolator(field):
        return ':%s' % field

    @staticmethod
    def row_ids(db, primary_key):
        return [db.cursor().lastrowid]


class DatabaseConnection(BaseConnection):

    DRIVER = sqlite3
    LEXICON = Lexicon

    @memoized_property
    def connect_args(self):
        return [self.db_config.database]

    @memoized_property
    def connect_kwargs(self):
        return {'isolation_level': 'DEFERRED'}



from future.standard_library import install_aliases
install_aliases()
from urllib.parse import urlparse
from collections import namedtuple
import pandas

from jardin.query_builders import \
    SelectQueryBuilder, \
    InsertQueryBuilder, \
    UpdateQueryBuilder, \
    DeleteQueryBuilder, \
    RawQueryBuilder
import jardin.config as config
from jardin.database.database_config import DatabaseConfig


class UnsupportedDriver(Exception): pass


class DatabaseConnections(object):

    _connections = {}
    _urls = {}

    SUPPORTED_SCHEMES = ('postgres', 'mysql', 'sqlite', 'snowflake')

    @classmethod
    def connection(self, db_name):
        if db_name not in self._connections:
            self._connections[db_name] = self.build_connection(db_name)
        return self._connections[db_name]

    @classmethod
    def build_connection(self, name):
        db = self.urls(name)
        if db.scheme not in self.SUPPORTED_SCHEMES:
            raise UnsupportedDriver('%s is not a supported driver' % db.scheme)
        elif db.scheme == 'postgres':
            import jardin.database.drivers.pg as driver
        elif db.scheme == 'mysql':
            import jardin.database.drivers.mysql as driver
        elif db.scheme == 'sqlite':
            import jardin.database.drivers.sqlite as driver
        elif db.scheme == 'snowflake':
            import jardin.database.drivers.sf as driver

        return driver.DatabaseConnection(db, name)


    @classmethod
    def urls(self, name):
        if len(self._urls) == 0:
            config.init()
        for (nme, url) in config.DATABASES.items():
            if url:
                self._urls[nme] = DatabaseConfig(url)
        return self._urls[name]


def set_defaults(func):
    def wrapper(self, *args, **kwargs):
        kwargs.update(
            model_metadata=self.model_metadata,
            scheme=self.db.db_config.scheme,
            lexicon=self.db.lexicon
            )
        return func(self, *args, **kwargs)
    return wrapper


class DatabaseAdapter(object):

    def __init__(self, db, model_metadata):
        self.db = db
        self.model_metadata = model_metadata

    @set_defaults
    def select(self, **kwargs):
        query = SelectQueryBuilder(**kwargs).query
        config.logger.debug(query)
        self.db.execute(*query)
        results = self.db.cursor().fetchall()
        return pandas.DataFrame(list(results), columns=self.columns())

    @set_defaults
    def write(self, query_builder, **kwargs):
        query = query_builder(**kwargs).query
        config.logger.debug(query)
        self.db.execute(*query)
        row_ids = self.db.lexicon.row_ids(self.db, kwargs['primary_key'])
        if len(row_ids) > 0:
            return self.select(where={kwargs['primary_key']: row_ids})

        return pandas.DataFrame(columns=self.columns())

    def insert(self, **kwargs):
        return self.write(InsertQueryBuilder, **kwargs)

    def update(self, **kwargs):
        return self.write(UpdateQueryBuilder, **kwargs)

    @set_defaults
    def delete(self, **kwargs):
        query = DeleteQueryBuilder(**kwargs).query
        config.logger.debug(query)
        self.db.execute(*query)

    @set_defaults
    def raw_query(self, **kwargs):
        query = RawQueryBuilder(**kwargs).query
        config.logger.debug(query)
        self.db.execute(*query)
        if self.db.cursor().description:
            results = self.db.cursor().fetchall()
            return pandas.DataFrame(list(results), columns=self.columns())
        else:
            return None

    def columns(self):
        cursor_desc = self.db.cursor().description
        columns = []
        if cursor_desc:
            columns = [col_desc[0] for col_desc in cursor_desc]
            if self.db.db_config.lowercase_columns:
                columns = [col.lower() for col in columns]
        return columns



from setuptools import setup
from os import path

import sys
if len(sys.argv) > 1 and sys.argv[1] == 'test':
      sys.jardin_setup = False
else:
      sys.jardin_setup = True

import jardin

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

setup(
      name = 'jardin',
      version = jardin.__version__,
      description = 'A dataframe-base ORM',
      long_description = long_description,
      url = 'https://github.com/instacart/jardin',
      author = 'Emmanuel Turlay',
      license = 'MIT',
      author_email = 'emmanuel@instacart.com',
      packages = ['jardin', 'jardin.database', 'jardin.database.drivers'],
      install_requires = [
      'pandas',
      'numpy',
      'psycopg2',
      'memoized_property',
      'inflect',
      'future'
      ],
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',
      test_suite='tests',
      classifiers = [
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: SQL',
      'Topic :: Database',
      'Topic :: Database :: Database Engines/Servers'
      ],
      keywords = 'postgres mysql database ORM'
      )


from memoized_property import memoized_property

import snowflake.connector as sf

from jardin.tools import retry
from jardin.database.drivers.pg import Lexicon as PGLexicon
from jardin.database.base import BaseConnection


class Lexicon(PGLexicon):

    @staticmethod
    def extrapolator(_):
        return '%s'

    @staticmethod
    def format_args(args):
        return args.values()


class DatabaseConnection(BaseConnection):

    DRIVER = sf
    LEXICON = Lexicon

    def __init__(self, db_config, name):
        super(DatabaseConnection, self).__init__(db_config, name)

    @memoized_property
    def connect_kwargs(self):
        return dict(
            user=self.db_config.username,
            password=self.db_config.password,
            account=self.db_config.account,
            warehouse=self.db_config.warehouse,
            database=self.db_config.database,
            schema=self.db_config.schema
            )

    @retry(sf.OperationalError, tries=3)
    def connect(self):
        return super(DatabaseConnection, self).connect()

    @retry(sf.InterfaceError, tries=3)
    def execute(self, *query):
        return super(DatabaseConnection, self).execute(*query)
